from skt.vault_utils import get_secrets


def get_hive_conn():
    from pyhive import hive

    hiveserver2 = get_secrets(path="ye/hiveserver2")
    host = hiveserver2["ip"]
    port = hiveserver2["port"]
    user = hiveserver2["user"]
    conn = hive.connect(host, port=port, username=user)
    return conn


def get_hdfs_conn():
    import os
    import pyarrow

    os.environ["ARROW_LIBHDFS_DIR"] = "/usr/hdp/3.0.1.0-187/usr/lib"
    conn = pyarrow.hdfs.connect(user="airflow")
    return conn


def get_pkl_from_hdfs(pkl_path):
    import pickle

    conn = get_hdfs_conn()
    byte_object = conn.cat(f"{pkl_path}")
    pkl_object = pickle.loads(byte_object)
    return pkl_object


def get_spark(scale=0, queue=None):
    import os
    from pyspark.sql import SparkSession
    import uuid

    tmp_uuid = str(uuid.uuid4())
    app_name = f"skt-{os.environ.get('USER', 'default')}-{tmp_uuid}"
    if not queue:
        if "JUPYTERHUB_USER" in os.environ:
            queue = "dmig_eda"
        else:
            queue = "airflow_job"
    os.environ["ARROW_PRE_0_15_IPC_FORMAT"] = "1"
    if scale in [1, 2, 3, 4]:
        spark = (
            SparkSession.builder.config("spark.app.name", app_name)
            .config("spark.driver.memory", f"{scale*8}g")
            .config("spark.executor.memory", f"{scale*3}g")
            .config("spark.executor.instances", f"{scale*8}")
            .config("spark.driver.maxResultSize", f"{scale*4}g")
            .config("spark.rpc.message.maxSize", "1024")
            .config("spark.yarn.queue", queue)
            .config("spark.ui.enabled", "false")
            .config("spark.port.maxRetries", "128")
            .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", "1")
            .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", "1")
            .enableHiveSupport()
            .getOrCreate()
        )
    else:
        spark = (
            SparkSession.builder.config("spark.app.name", app_name)
            .config("spark.driver.memory", "6g")
            .config("spark.executor.memory", "8g")
            .config("spark.shuffle.service.enabled", "true")
            .config("spark.dynamicAllocation.enabled", "true")
            .config("spark.dynamicAllocation.maxExecutors", "200")
            .config("spark.driver.maxResultSize", "6g")
            .config("spark.rpc.message.maxSize", "1024")
            .config("spark.yarn.queue", queue)
            .config("spark.ui.enabled", "false")
            .config("spark.port.maxRetries", "128")
            .config("spark.executorEnv.ARROW_PRE_0_15_IPC_FORMAT", "1")
            .config("spark.yarn.appMasterEnv.ARROW_PRE_0_15_IPC_FORMAT", "1")
            .enableHiveSupport()
            .getOrCreate()
        )
    spark.conf.set("spark.sql.execution.arrow.enabled", "true")
    return spark


def hive_execute(query):
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute(query)
    c.close()
    conn.close()


def hive_get_result(query):
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute(query)
    result = c.fetchall()
    c.close()
    conn.close()
    return result


def hive_to_pandas(query):
    import uuid

    tmp_id = str(uuid.uuid4()).replace("-", "_")
    ctas = f"CREATE TABLE dumbo.{tmp_id} stored as parquet as {query}"
    conn = get_hive_conn()
    c = conn.cursor()
    c.execute("set parquet.column.index.access=false")
    c.execute(ctas)
    c.close()
    conn.close()
    table_path = f"/warehouse/tablespace/managed/hive/dumbo.db/{tmp_id}/delta_0000001_0000001_0000"
    return parquet_to_pandas(table_path)


def parquet_to_pandas(hdfs_path):
    from pyarrow import parquet

    hdfs = get_hdfs_conn()
    df = parquet.read_table(hdfs_path, filesystem=hdfs).to_pandas()
    df.info()
    return df


def pandas_to_parquet(pandas_df, hdfs_path, spark):
    df = spark.createDataFrame(pandas_df)
    df.write.mode("overwrite").parquet(hdfs_path)


def slack_send(
    text="This is default text", username="SKT", channel="#leavemealone", icon_emoji=":large_blue_circle:", blocks=None,
):
    import requests
    from skt.vault_utils import get_secrets

    token = get_secrets("slack")["bot_token"]["airflow"]
    proxy = get_secrets("proxy")["proxy"]
    proxies = {
        "http": proxy,
        "https": proxy,
    }
    headers = {
        "Content-Type": "application/json;charset=utf-8",
        "Authorization": f"Bearer {token}",
    }
    json_body = {
        "username": username,
        "channel": channel,
        "text": text,
        "blocks": blocks,
        "icon_emoji": icon_emoji,
    }
    r = requests.post("https://www.slack.com/api/chat.postMessage", proxies=proxies, headers=headers, json=json_body,)
    r.raise_for_status()
    if not r.json()["ok"]:
        raise Exception(r.json())


def get_github_util():
    from skt.github_utils import GithubUtil

    github_token = get_secrets("github/sktaiflow")["token"]
    proxy = get_secrets("proxy")["proxy"]
    proxies = {
        "http": proxy,
        "https": proxy,
    }
    g = GithubUtil(github_token, proxies)
    return g



from pathlib import Path
from typing import List, Dict, Any
from enum import Enum

import requests
import json
import subprocess
import shlex
import os
import tarfile
import shutil
import re
from joblib import dump
from datetime import datetime

import lightgbm
import xgboost

from skt.vault_utils import get_secrets


MLS_MODEL_DIR = os.path.join(Path.home(), "mls_temp_dir")
MODEL_BINARY_NAME = "model.joblib"
MODEL_TAR_NAME = "model.tar.gz"
MODEL_META_NAME = "model.json"
S3_DEFAULT_PATH = "s3a://mls-model-registry"

EDD_OPTIONS = """-Dfs.s3a.proxy.host=awsproxy.datalake.net \
                 -Dfs.s3a.proxy.port=3128 \
                 -Dfs.s3a.endpoint=s3.ap-northeast-2.amazonaws.com \
                 -Dfs.s3a.security.credential.provider.path=jceks:///user/tairflow/s3_mls.jceks \
                 -Dfs.s3a.fast.upload=true -Dfs.s3a.acl.default=BucketOwnerFullControl"""


HEADER = {"Content-Type": "application/json"}
AB_URL_PREFIX = "https://ab-internal"
MLS_META_API_URL = "/api/v1/meta"


def set_model_name(comm_db, params):
    secret = get_secrets("mls")
    if comm_db[-3:] == "dev":  # stg
        url = secret["dev_url"]
    else:  # prd
        url = secret["prd_url"]
    requests.post(url, data=json.dumps(params))


def get_recent_model_path(comm_db, model_key):
    secret = get_secrets("mls")
    if comm_db[-3:] == "dev":  # stg
        url = secret["dev_url"]
    else:  # prd
        url = secret["prd_url"]
    return requests.get(f"{url}/latest").json()[model_key]


def get_model_name(key):
    secret = get_secrets("mls")
    url = secret["prd_url"]
    return requests.get(f"{url}/latest").json()[key]


class ModelLibrary(Enum):
    LIGHTGBM = "lightgbm"
    XGBOOST = "xgboost"


class AWSENV(Enum):
    STG = "stg"
    PRD = "prd"
    DEV = "dev"


class MLSModelError(Exception):
    def __init__(self, msg):
        super().__init__(msg)


def save_model(
    model,
    model_name: str,
    model_version: str,
    aws_env: AWSENV = AWSENV.STG.value,
    feature_list: List[str] = None,
    force: bool = False,
    edd: bool = False,
) -> None:
    """
    Upload model_binary(model.tar.gz) / model_meta(model.json) to AWS S3 from YE
    Args. :
        - model         :   Model object generated by LightGBM or XGBoost
        - model_name    :   (str) Model name to be saved on S3 and MLS
        - model_version :   (str) Model version to be saved on S3 ans MLS
        - aws_env       :   (str) AWS ENV in 'stg / prd / dev' (default is 'stg')
        - feature_list  :   (List[str]) List of features used in ML Model in string type (only for XGBoost model type)
        - force         :   (bool) Force to overwrite model files on S3 if exists (default is False)
        - edd           :   (bool) True if On-prem env is on EDD (default is False)
    """

    assert type(model_name) == str
    assert type(model_version) == str

    if not bool(re.search("^[A-Za-z0-9_]+_model$", model_name)):
        raise MLSModelError(
            "model_name should follow naming rule. MUST be in alphabet, number, underscore and endwiths '_model'"
        )

    if not bool(re.search("^[A-Za-z0-9_]+$", model_version)):
        raise MLSModelError("model_name should follow naming rule. MUST be in alphabet, number, underscore")

    def check_model_library(model):
        if isinstance(model, lightgbm.Booster):
            return ModelLibrary.LIGHTGBM.value, lightgbm.__version__
        elif isinstance(model, xgboost.XGBModel):
            return ModelLibrary.XGBOOST.value, xgboost.__version__
        else:
            raise MLSModelError("Input Model is none of LightGBM or XGBoost type")

    def get_feature_list(model, model_library: str):
        if model_library == ModelLibrary.LIGHTGBM.value:
            return model.feature_name()
        elif model_library == ModelLibrary.XGBOOST.value:
            if all(isinstance(s, str) for s in feature_list) and len(feature_list) == len(model.feature_importances_):
                return feature_list
            else:
                raise MLSModelError(
                    "Input feature list is not matching the number of model input feature or list STRING type"
                )

    model_library, model_library_version = check_model_library(model)
    feature_list = get_feature_list(model, model_library)

    s3_path = S3_DEFAULT_PATH
    if aws_env in (AWSENV.STG.value, AWSENV.PRD.value):
        s3_path = f"{S3_DEFAULT_PATH}-{aws_env}"
    s3_path = f"{s3_path}/{model_name}/{model_version}"

    model_meta = {
        "name": model_name,
        "version": model_version,
        "model_lib": model_library,
        "model_lib_version": model_library_version,
        "model_data": f"{s3_path}/{MODEL_TAR_NAME}",
        "features": feature_list,
    }

    model_path = os.path.join(
        MLS_MODEL_DIR, f"{aws_env}_{model_name}_{model_version}_{datetime.today().strftime('%Y%m%d_%H%M%S')}",
    )
    try:
        if not os.path.exists(os.path.join(model_path, MODEL_BINARY_NAME)):
            if not os.path.exists(model_path):
                os.makedirs(model_path)

            dump(model, os.path.join(model_path, MODEL_BINARY_NAME))
            with tarfile.open(f"{model_path}/{MODEL_TAR_NAME}", "w") as tar:
                tar.add(f"{model_path}/{MODEL_BINARY_NAME}")

            with open(os.path.join(model_path, MODEL_META_NAME), "w") as f:
                json.dump(model_meta, f)
        else:
            raise MLSModelError(f"{model_name} / {model_version} is already in PATH ({model_path})")

        cmd_mkdir = f"hdfs dfs {EDD_OPTIONS if edd else ''} -mkdir -p {s3_path}"
        cmd_load_model_to_s3 = f"hdfs dfs {EDD_OPTIONS if edd else ''} -put {'-f' if force else ''} \
            {os.path.join(model_path, MODEL_TAR_NAME)} {s3_path}"
        cmd_load_meta_to_s3 = f"hdfs dfs {EDD_OPTIONS if edd else ''} -put {'-f' if force else ''} \
            {os.path.join(model_path, MODEL_META_NAME)} {s3_path}"

        process_mkdir = subprocess.Popen(shlex.split(cmd_mkdir), stdout=subprocess.PIPE, stdin=subprocess.PIPE,)
        process_mkdir.wait()
        if process_mkdir.returncode != 0:
            raise MLSModelError(f"Making Directory on S3 ({s3_path}) is FAILED")

        process_model_binary = subprocess.Popen(
            shlex.split(cmd_load_model_to_s3), stdout=subprocess.PIPE, stdin=subprocess.PIPE,
        )
        process_model_binary.wait()
        if process_model_binary.returncode != 0:
            raise MLSModelError(
                f"Load model_binary(model.tar.gz) to S3 ({s3_path}) is FAILED. (Maybe Already exists on S3)"
            )

        process_model_meta = subprocess.Popen(
            shlex.split(cmd_load_meta_to_s3), stdout=subprocess.PIPE, stdin=subprocess.PIPE,
        )
        process_model_meta.wait()
        if process_model_meta.returncode != 0:
            raise MLSModelError(f"Load model_meta(meta.json) to S3 ({s3_path}) is FAILED")

    finally:
        shutil.rmtree(model_path, ignore_errors=True)


def get_meta_table(meta_table: str, aws_env: AWSENV = AWSENV.STG.value) -> Dict[str, Any]:
    """
    Get a meta_table information
    Args. :
        - meta_table   :   (str) the name of meta_table
        - aws_env      :   (str) AWS ENV in 'stg / prd / dev' (default is 'stg')
    Returns :
        - Dictionary value of meta_table (id / name / description / schema / items / created_at / updated_at)
    """
    assert type(meta_table) == str
    assert type(aws_env) == str

    url = AB_URL_PREFIX
    if aws_env in (AWSENV.STG.value, AWSENV.DEV.value):
        url = f"{url}.{aws_env}"
    url = f"{url}.sktmls.com{MLS_META_API_URL}/{meta_table}"

    response = requests.get(url).json()
    results = response.get("results")

    if not results:
        raise MLSModelError(response.get("error"))
    else:
        return results


def create_meta_table_item(
    meta_table: str, item_name: str, item_dict: Dict[str, Any], aws_env: AWSENV = AWSENV.STG.value
) -> None:
    """
    Create a meta_item
    Args. :
        - meta_table   :   (str) the name of meta_table
        - item_name    :   (str) the name of meta_item to be added
        - item_dict    :   (dict) A dictionary type (item-value) value to upload to or update of the item
        - aws_env      :   (str) AWS ENV in 'stg / prd / dev' (default is 'stg')
        - force        :   (bool) Force to overwrite(update) the item_meta value if already exists
    """
    assert type(meta_table) == str
    assert type(item_name) == str
    assert type(item_dict) == dict
    assert type(aws_env) == str

    meta_table_info = get_meta_table(meta_table, aws_env)

    values_data = dict()
    for field_name, field_spec in meta_table_info["schema"].items():
        values_data[field_name] = item_dict.get(field_name)

    request_data = dict()
    request_data["name"] = item_name
    request_data["values"] = values_data

    url = AB_URL_PREFIX
    if aws_env in (AWSENV.STG.value, AWSENV.DEV.value):
        url = f"{url}.{aws_env}"
    url = f"{url}.sktmls.com{MLS_META_API_URL}/{meta_table}/items"

    response = requests.post(url, headers=HEADER, data=json.dumps(request_data)).json()
    results = response.get("results")

    if not results:
        raise MLSModelError(response.get("error"))


def update_meta_table_item(
    meta_table: str, item_name: str, item_dict: Dict[str, Any], aws_env: AWSENV = AWSENV.STG.value
) -> None:
    """
    Update a meta_item
    Args. :
        - meta_table   :   (str) the name of meta_table
        - item_name    :   (str) the name of meta_item to be added
        - item_dict    :   (dict) A dictionary type (item-value) value to upload to or update of the item
        - aws_env      :   (str) AWS ENV in 'stg / prd / dev' (default is 'stg')
    """
    assert type(meta_table) == str
    assert type(item_name) == str
    assert type(item_dict) == dict
    assert type(aws_env) == str

    meta_table_info = get_meta_table(meta_table, aws_env)

    values_data = dict()
    for field_name, field_spec in meta_table_info["schema"].items():
        values_data[field_name] = item_dict.get(field_name)

    request_data = dict()
    request_data["name"] = item_name
    request_data["values"] = values_data

    url = AB_URL_PREFIX
    if aws_env in (AWSENV.STG.value, AWSENV.DEV.value):
        url = f"{url}.{aws_env}"
    url = f"{url}.sktmls.com{MLS_META_API_URL}/{meta_table}/items/{item_name}"

    response = requests.put(url, headers=HEADER, data=json.dumps(request_data)).json()
    results = response.get("results")

    if not results:
        raise MLSModelError(response.get("error"))


def get_meta_table_item(meta_table: str, item_name: str, aws_env: AWSENV = AWSENV.STG.value) -> Dict[str, Any]:
    """
    Get a meta_table information
    Args. :
        - meta_table   :   (str) the name of meta_table
        - item_name    :   (str) the name of meta_item to be added
        - aws_env      :   (str) AWS ENV in 'stg / prd / dev' (default is 'stg')
    Returns :
        - A dictionary type (item-value) value of the item_meta
    """
    assert type(meta_table) == str
    assert type(item_name) == str
    assert type(aws_env) == str

    url = AB_URL_PREFIX
    if aws_env in (AWSENV.STG.value, AWSENV.DEV.value):
        url = f"{url}.{aws_env}"
    url = f"{url}.sktmls.com{MLS_META_API_URL}/{meta_table}/items/{item_name}"

    response = requests.get(url).json()
    results = response.get("results")

    if not results:
        raise MLSModelError(response.get("error"))
    else:
        return results



import requests
from skt.vault_utils import get_secrets


DATA_CATALOG_SECRETS_NAME = "data_catalog"


def get_sources():
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/sources").json()


def get_source(source):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/sources/{source}").json()


def get_tables(source, limit=1000):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/sources/{source}/tables?limit={limit}").json()


def get_columns(source, table_id):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/sources/{source}/tables/{table_id}/columns").json()


def search_table_by_name(name):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/v1/search/tables?name={name}").json()


def search_table_by_dc(description):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/v1/search/tables?description={description}").json()


def search_column_by_dc(description):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/v1/search/columns?description={description}").json()


def search_column_by_name(name):
    secrets = get_secrets(DATA_CATALOG_SECRETS_NAME)
    return requests.get(f"{secrets['url_prd']}/v1/search/columns?name={name}&fuzziness=0").json()
